
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 2 of the License, or
#       (at your option) any later version.
#       
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#       
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.


# author: Syrischer Raubelefant
# soaplogic - logical part of soap - sun of a peach

import sys
import math
import platform

import pygame

__name__ = 'soaplogic'


### Math stuff ##############################

class Vector(object):
    
    def __init__(self, x, y):
        super(Vector, self).__init__()
        self.x = float(x)
        self.y = float(y)
    
    def __sub__(self, cp):
        return Vector(self.x - cp.x, self.y - cp.y)
    
    def __add__(self, cp):
        return Vector(self.x + cp.x, self.y + cp.y)
    
    def __mul__(self, cp):
        if isinstance(cp, Vector):
            return self.dot_product(cp)
        else:
            return Vector(self.x * cp, self.y * cp)
    
    def dot_product(self, cp):
        return self.x * cp.x + self.y * cp.y
    
    def get_length(self):
        return math.sqrt(self.x * self.x + self.y * self.y)
    
    def normalize(self):
        l = self.length
        if l == 0:
            raise ValueError('cannot normalize zero vector')
        else:
            self.x /= l
            self.y /= l
    
    def set_length(self, length):
        self.normalize()
        self.x *= length
        self.y *= length
    
    length = property(get_length, set_length)
    
    def angle_to(self, cp):
        angle = math.acos((self * cp) / (self.length * cp.length))
        return math.degrees(angle)
    
    def negate(self):
        self.x *= -1
        self.y *= -1
    
    def copy(self):
        return Vector(self.x, self.y)


def vec_from_to(p1, p2):
    return Vector(p2[0] - p1[0], p2[1] - p1[1])

def calc_distance(p1, p2):
    v = vec_from_to(p1, p2)
    return v.length

def is_in_between(x, a, b):
    if (a <= x <= b) or (a >= x >= b):
        return True
    else:
        return False

def calc_values_in_between(a, b):
    d = b - a
    if d != 0:
        # -> step will be either 1 or -1
        step = d / abs(d)
        return list( range(a + step, b, step) )
    
#############################################

### Timing stuff ############################

_global_clock = None

# use this before you use any other timing functions
def init_timing():
    global _global_clock
    _global_clock = pygame.time.Clock() 

# update once per frame
def update_timing(fps=0):
    global _global_clock
    _global_clock.tick(fps)

# time passed since last update (in seconds - float)
def time_passed():
    global _global_clock
    return float(_global_clock.get_time() / 1000.0)

def time_passed_ms():
    global _global_clock
    return _global_clock.get_time()

def get_fps():
    global _global_clock
    return _global_clock.get_fps()

#############################################


def errmsg(msg):
    maj, min, pat = platform.python_version_tuple()
    if maj == '2':
        print 'error: '+msg
    elif maj == '3':
        print('error: '+msg)
    sys.exit()


def readlines(filename):
    f = open(filename)
    lines = f.readlines()
    f.close()
    
    # remove all blank lines
    def f(x):
        return not x.isspace()
    
    f_lines = filter(f, lines)
    del lines
    return f_lines


class Identifiable(object):
    
    def __init__(self):
        super(Identifiable, self).__init__()
        self.id = None


class Coord2D(object):
    
    def __init__(self):
        super(Coord2D, self).__init__()
        self.x = 0.0
        self.y = 0.0
    
    def get_pos(self):
        return (self.x, self.y)
    
    def set_pos(self, new_pos):
        self.x, self.y = new_pos
        
    # pos is supposed to be a tuple of the form (x, y)
    pos = property(get_pos, set_pos)


class Tile(Identifiable):
    
    def __init__(self, id=None, solid=False):
        super(Tile, self).__init__()
        # id determines how to render it later
        self.id = id
        self.solid = solid
    
    def action(self): # dummy
        pass


class TileMap(object):
    
    def __init__(self, map_size, tile_size):
        super(TileMap, self).__init__()
        self.size = map_size
        self.tile_size = tile_size
        self.real_size = self.calc_real_size()
        self.generate_empty_tilemap()
    
    def calc_real_size(self):
        w, h = self.size
        t_w, t_h = self.tile_size
        return (w * t_w, h * t_h)
    
    def get_tile(self, tile_pos):
        assert self.is_valid_index(tile_pos), 'get_tile: invalid index'
        x, y = tile_pos
        return self.tiles[y][x]
    
    def set_tile_prop(self, tile_pos, id, solid):
        x, y = tile_pos
        self.tiles[y][x].id = id
        self.tiles[y][x].solid = solid
    
    def generate_empty_tilemap(self):
        w, h = self.size
        self.tiles = [ [Tile() for x in range(w)] for y in range(h) ]

    def tile_index_to_coords(self, tile_pos):
        t_w, t_h = self.tile_size
        x, y = tile_pos
        return (float(x * t_w), float(y * t_h))
    
    def coords_to_tile_index(self, pos):
        t_w, t_h = self.tile_size
        x, y = pos
        return (int(x / t_w), int(y / t_h)) # same as: math.trunc()
    
    def is_valid_pos(self, pos):
        tile_pos = self.coords_to_tile_index(pos)
        return self.is_valid_index(tile_pos)
    
    def is_valid_index(self, tile_pos):
        x, y = tile_pos
        w, h = self.size
        if is_in_between(x, 0, w-1) and is_in_between(y, 0, h-1):
            return True
        else:
            return False
    
    def aligend_to_tilegrid(self, pos):
        t_w, t_h = self.tile_size
        x, y = pos
        if x % t_w == 0 and y % t_h == 0:
            return True
        else:
            return False


# file format of tilemaps (*.sm) is:
# ->soap_map<-
# w h tile_w tile_h
# id|# id|_ id|# id|#
# id|# id|_ id|# id|_
# id|# id|_ id|_ id|_
# id|# id|# id|# id|_
# [...]

class TileMapLoader(object):
    
    MAGIC = '->soap_map<-'
    
    def __init__(self, filename):
        super(TileMapLoader, self).__init__()
        self.filename = filename
        self.tilemap = None
        self.load(filename)
    
    def parse_tile_entry(self, entry):
        id, solid = entry.split('|')
        if solid == '#':
            solid = True
        elif solid == '_':
            solid = False
        else: # in case someone messed it up
            errmsg('the solid flag has to be either # or _')
        return (id, solid)
    
    # TileMapLoader.load() creates a TileMap from file and returns it
    def load(self, filename):
        lines = readlines(filename)

        # make sure, the file containts at least the magic & size informations
        assert len(lines) > 2, filename+' is broken'
        
        # check the magic word
        magic_line = lines.pop(0)
        assert TileMapLoader.MAGIC in magic_line, filename+' misses magic'
        
        # get the tilemaps size informations
        size_line = lines.pop(0)
        size_list = [int(entry) for entry in size_line.split()]
        assert len(size_list) == 4, 'invalid tilemap size information'
        w, h, tile_w, tile_h = size_list
        
        assert len(lines) >= h, 'invalid tilemap data'
        # the resulting tilemap goes here
        self.tilemap = TileMap((w, h), (tile_w, tile_h))
        
        # parse the tilemap data
        for y, line in enumerate(lines):
            entries = line.split()
            for x, entry in enumerate(entries):
                id, solid = self.parse_tile_entry(entry)
                self.tilemap.set_tile_prop((x, y), id, solid)


class Item(Identifiable, Coord2D):
    
    def __init__(self):
        super(Item, self).__init__()
        self.taken = False
        self.value = None
        self.state = State.DEFAULT
    
    def _action(self):
        self.taken = True
        self.action()
    
    def action(self): # callback function
        pass


class Figure(Identifiable, Coord2D):
    
    def __init__(self):
        super(Figure, self).__init__()
        
        self.prev_pos = self.pos
        
        self.state = State.STANDING
        self.alive = True
        
        # speed in pixels per second
        self.speed = 0.0
        # tuple (x, y)
        self.dest_pos = None
        self.patrol_route = None
        self.route_index = 0
    
    # route is a list of tuples (x, y)
    def patrol(self, route, speed):
        assert len(route) >= 2, 'patrol: route too short'
        self.route_index = 0
        self.patrol_route = route
        self.speed = speed
        
    def on_death(self): # callback function
        pass
    
    def _die(self): # internal function - do not use it
        if self.alive:
            self.alive = False
            self.state = State.DEAD
            self.on_death()
    
    # go_to & update are "raw" functions. The caller has to take care of the
    # destination coordinates
    def go_to(self, dest, speed):
        self.dest_pos = dest
        self.speed = float(speed)
    
    def is_at_destination(self):
        return (self.dest_pos == self.pos)
    
    def has_valid_destination(self):
        if self.dest_pos is None:
            return False
        else:
            return True
    
    def update_movement_state(self):
        if self.has_valid_destination() == False and self.pos == self.prev_pos:
            self.state = State.STANDING
        else:
            self.state = State.MOVING
    
    def update_pos(self):
        # destination has been set
        # when destination is reached, destination becomes invalid (reset)
        if self.has_valid_destination():
            vec = vec_from_to(self.pos, self.dest_pos)
            distance = vec.length
            single_step = self.speed * time_passed()
            if distance < single_step: # dest_pos almost reached 
                single_step = distance
            vec.length = single_step
            self.add_vector(vec)
            if self.is_at_destination():
                self._on_reach_dest()
    
    def update(self):
        if self.alive:
            self.prev_pos = self.pos # save old posisiton
            self.update_route()
            self.update_pos()
            self.update_movement_state()
        else:
            self.patrol_route = None
        
    def add_vector(self, v):
        self.x += v.x
        self.y += v.y
    
    def is_at_route_node(self):
        return (self.patrol_route is not None and
                self.has_valid_destination() == False)
    
    def update_route(self):
        if self.is_at_route_node():
            self.route_index += 1
            if self.route_index >= len(self.patrol_route):
                self.route_index = 1
                self.patrol_route.reverse()
            dest = self.patrol_route[self.route_index]
            self.go_to(dest, self.speed)

    def _on_reach_dest(self):
        self.dest_pos = None # reset destination

class Direction(object):
    
    def __init__(self):
        super(Direction, self).__init__()
    
    LEFT = 0
    RIGHT = 1
    UP = 2
    DOWN = 3


class State(object):
    
    def __init__(self):
        super(State, self).__init__()
    
    DEFAULT = 0
    STANDING = 1
    MOVING = 2
    DEAD = 3


# file format of level files (*.sl) is:
# ->soap_level<-
# m soapmap.sm
# p x y speed_float 1x1 10x1 10x10 5x10
# e x y
# e x y
# [...]
# i x y
# i x y


class Game(object):
    
    MAGIC = '->soap_level<-'
    
    def __init__(self):
        super(Game, self).__init__()
        init_timing()
        
        self.score = 0
        
        self.player = None
        self.enemies = []
        self.items = []
        
        self.map = None
    
    def reset(self):
        self.score = 0
        
        del self.player
        self.player = None
        
        del self.enemies[:]
        del self.items[:]
        
        del self.map
        self.map = None
        
    def on_finish(self): # callback function
        pass
    
    def calc_intermediate_indices(self, i1, i2):
        i1_x, i1_y = i1 # from
        i2_x, i2_y = i2 # to
        
        d_x = i2_x - i1_x
        d_y = i2_y - i1_y
        
        indices = []
        
        if d_x != 0 and d_y == 0:
            xs = calc_values_in_between(i1_x, i2_x)
            for x in xs:
                indices.append((x, i1_y)) # i1_y == i2_y
        elif d_y != 0 and d_x == 0:
            ys = calc_values_in_between(i1_y, i2_y)
            for y in ys:
                indices.append((i1_x, y)) # i1_x == i2_x
        else: 
            s = 'route: cannot move diagonally'+ str(i1) + str(i2)
            errmsg(s)
        return indices
    
    def calc_full_route(self, route):
        prev_pos = None
        final_route = []
        for tile_pos in route:
            if prev_pos is not None:
                final_route.append(prev_pos)
                
                new_indices = self.calc_intermediate_indices(prev_pos, tile_pos)
                final_route.extend(new_indices)
                
            # save index
            prev_pos = tile_pos
        final_route.append(prev_pos) # prev_pos is actually pos
        return final_route
    
    def valid_routing_indices(self, route):
        final_route = self.calc_full_route(route)
        for tile_pos in final_route:
            # TileMap.get_tile checks for valid index
            tile = self.map.get_tile(tile_pos)
            
            if tile.solid == True:
                return False
        
        return True
    
    # route is a list of tile indices
    def patrol_figure(self, figure, route, speed):
        if self.valid_routing_indices(route):
            routing_coords = []
            for tile_pos in route:
                pos = self.map.tile_index_to_coords(tile_pos)
                routing_coords.append(pos)
            figure.patrol(routing_coords, speed)
        else:
            errmsg('patrol_figure: invalud routing indices (solid?) ' + str(route))
    
    def move_figure(self, fig, dir, speed):
        if self.map.aligend_to_tilegrid(fig.pos): # covers a tile 
            i_x, i_y = self.map.coords_to_tile_index(fig.pos)
            if dir == Direction.LEFT:
                i_x -= 1
            elif dir == Direction.RIGHT:
                i_x += 1
            elif dir == Direction.UP:
                i_y -= 1
            elif dir == Direction.DOWN:
                i_y += 1
            else:
                errmsg('move_figure: invalid direction')
            dest_tile_pos = (i_x, i_y)
            
            # still inside the tilemap & not solid ?
            if self.map.is_valid_index(dest_tile_pos):
                tile = self.map.get_tile(dest_tile_pos)
                if tile.solid == False:
                    dest = self.map.tile_index_to_coords(dest_tile_pos)
                    fig.go_to(dest, speed)
    
    def move_player(self, dir, speed):
        self.move_figure(self.player, dir, speed)
    
    def collides(self, a, b):
        t_w, t_h = self.map.tile_size
        r_a = pygame.Rect(a.x, a.y, t_w, t_h)
        r_b = pygame.Rect(b.x, b.y, t_w, t_h)
        return r_a.colliderect(r_b)
    
    def player_collides(self, cp):
        return self.collides(self.player, cp)
    
    def update_items(self):
        if len(self.items) > 0:
            remaining_items = []
            for i in self.items:
                if self.collides(self.player, i):
                    self.score += i.value
                    i._action()
                else:
                    remaining_items.append(i)
            # some collisions accoured
            if self.items != remaining_items:
                del self.items
                self.items = remaining_items
            
            if not self.items:
                self.on_finish()
        
    
    def update_enemies(self):
        for e in self.enemies:
            e.update()
            
            if self.collides(self.player, e):
                self.player._die()
        
    def update(self):
        update_timing()
        self.player.update()
        self.update_enemies()
        self.update_items()
    
    def load_level(self, filename):
        self.reset()
        lines = readlines(filename)
        
        assert len(lines) > 1, 'load_level: invalid level file'
        # check the magic word
        magic_line = lines.pop(0)
        assert Game.MAGIC in magic_line, filename+' misses magic'
        
        for line in lines:
            words = line.split()
            a = words.pop(0)
            if a == 'm': # map: m filename.sm
                mapname = words.pop(0)
                self.load_map(mapname)
            elif a in 'pe': # p/e id x y speed <patrol path XxY>
                assert len(words) >= 3, 'no id/coords in '+filename
                id = words.pop(0)
                x = int(words.pop(0))
                y = int(words.pop(0))
                
                fig = Figure()
                
                if a == 'p':
                    self.register_player(fig, id, (x, y))
                else: # a == 'e'
                    self.register_enemy(fig, id, (x, y))
                
                # patrolling requires at least 3 more words:
                # 1 = speed, 2 & 3 = two points (minimalistic config.)
                if len(words) >= 3:
                    # patrolling path available
                    speed = float(words.pop(0))
                    patrol_route = []
                    # words now contains a list of indices for patrolling
                    for entry in words: # encoded as 1x2 for x=1 and y=2
                        str_x, str_y = entry.split('x')
                        i_x = int(str_x)
                        i_y = int(str_y)
                        patrol_route.append((i_x, i_y))
                    self.patrol_figure(fig, patrol_route, speed)
                    
            elif a == 'e': # enemy
                id = words.pop(0)
                x = int(words.pop(0))
                y = int(words.pop(0))
                self.register_enemy(Figure(), id, (x, y))
            elif a == 'i':
                id = words.pop(0)
                value = int(words.pop(0))
                x = int(words.pop(0))
                y = int(words.pop(0))
                self.register_item(Item(), id, value, (x, y))
            else:
                errmsg('load_level: invalid level file ('+filename+') -> '+a)

    def create_empty_map(self, map_size, tile_size):
        self.map = TileMap(map_size, tile_size)
    
    def load_map(self, filename):
        self.map = TileMapLoader(filename).tilemap
        
    def register_player(self, p, id, tile_pos):
        assert self.map.is_valid_index(tile_pos), 'register: invalid tile index'
        self.player = p
        self.player.pos = self.map.tile_index_to_coords(tile_pos)
        self.player.id = id
    
    def register_enemy(self, e, id, tile_pos):
        assert e not in self.enemies, 'cannot register same enemy twice'
        assert self.map.is_valid_index(tile_pos), 'register: invalid tile index'
        e.pos = self.map.tile_index_to_coords(tile_pos)
        e.id = id
        self.enemies.append(e)
    
    def register_item(self, i, id, value, tile_pos):
        assert i not in self.items, 'cannot register same item twice'
        assert self.map.is_valid_index(tile_pos), 'register: invalid tile index'
        i.pos = self.map.tile_index_to_coords(tile_pos)
        i.id = id
        i.value = value
        self.items.append(i)


class Menu(object):
    
    def __init__(self):
        super(Menu, self).__init__()
        self.entries = []
        self.active_entry = 0
    
    def execute(self):
        entry = self.entries[self.active_entry]
        entry.action()
    
    def next_entry(self):
        self.active_entry += 1
        if self.active_entry >= len(self.entries):
            self.active_entry = 0 # first entry
    
    def previous_entry(self):
        self.active_entry -= 1
        if self.active_entry < 0:
            self.active_entry = len(self.entries) - 1 # last entry
    
    def add_entry(self, text, cb):
        entry = MenuEntry()
        entry.text = text
        entry.action = cb
        self.entries.append(entry)


class MenuEntry(object):
    
    def __init__(self):
        super(MenuEntry, self).__init__()
        self.text = None
    
    def action(self): # callback function
        pass
