
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 2 of the License, or
#       (at your option) any later version.
#       
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#       
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.


# author: Syrischer Raubelefant
# soaprenderer - the renderer of soap - sun of a peach

import pygame
import soaplogic
import copy

__name__ = 'soaprenderer'


# file format of animation sets (*.sas) is:
# ->soap_animation_set<-
# state delay id0 id1 id2 id3[...]
# state delay id49 id50
# [...]

###############################################################################

# file format of rendering properties (*.sr) is:
# ->soap_renderer<-
# % id0 anim_set0.sas
# % id1 anim_set1.sas
# % id2 anim_set2.sas
# - id3 tile_filename0.png
# - id4 tile_filename1.png
# - id5 tile_filename2.png
# [...]

class IDManager(object):
    
    # stuff for *.sas files
    MAGIC_SAS = '->soap_animation_set<-'
    
    # stuff for *.sr files
    MAGIC_SR = '->soap_renderer<-'
    C_IMG = '-'
    C_AS = '%'
    
    def __init__(self, sr_filename):
        super(IDManager, self).__init__()
        
        # images
        self.img_dict = {}
        # animation sets
        self.as_dict = {}
        self.load_rendering_properties(sr_filename)
    
    def register_image(self, filename, id):
        if id not in self.img_dict:
            img = self.load_image(filename)
            self.img_dict[id] = img
    
    def register_animation_set(self, filename, id):
        if id not in self.as_dict:
            anim_set = self.load_animation_set(filename)
            self.as_dict[id] = anim_set
    
    # pure image loading routine, no checks
    def load_image(self, filename):
        img = pygame.image.load(filename)
        converted_image = img.convert_alpha()
        del img
        return converted_image
    
    def parse_state(self, state_str):
        state = None
        if state_str == 'DEFAULT':
            state = soaplogic.State.DEFAULT
        elif state_str == 'STANDING':
            state = soaplogic.State.STANDING
        elif state_str == 'MOVING':
            state = soaplogic.State.MOVING
        elif state_str == 'DEAD':
            state = soaplogic.State.DEAD
        else:
            errmsg('parse_state(): invalid state_str '+state_str)
        return state
    
    def parse_sas_line(self, line):
        words = line.split()
        # make sure that state, delay and at least one frameID are present
        assert len(words) > 2, 'error while parsing animation set file'
        state_str = words.pop(0)
        delay = int(words.pop(0))
        
        state = self.parse_state(state_str)
        
        # words now contains a list of id's (frames)
        anim = Animation()
        for frame_id in words:
            anim.add_frame_id(frame_id)
        anim.delay = delay
        
        return (state, anim)
        
    def load_animation_set(self, filename):
        animation_set = {}
        lines = soaplogic.readlines(filename)
        
        assert len(lines) > 1, filename+' is broken'
        # check the magic word
        magic_line = lines.pop(0)
        assert IDManager.MAGIC_SAS in magic_line, filename+' misses magic'
        
        for line in lines:
            state, anim = self.parse_sas_line(line)
            animation_set[state] = anim
        return animation_set
    
    def parse_sr_line(self, line):
        words = line.split()
        assert len(words) == 3, 'error while parsing rendering properties'
        
        a, id, filename = words
        return (a, id, filename)
    
    # loads all the animation sets/images corresponding to every individual id
    def load_rendering_properties(self, filename):
        lines = soaplogic.readlines(filename)
        
        assert len(lines) > 1, filename+' is broken'
        # check the magic word
        magic_line = lines.pop(0)
        assert IDManager.MAGIC_SR in magic_line, filename+' misses magic'
        
        for line in lines:
            a, id, filename = self.parse_sr_line(line)

            if a == IDManager.C_IMG:
                self.register_image(filename, id)
            elif a == IDManager.C_AS:
                self.register_animation_set(filename, id)
            else:
                errmsg('invalid control character in '+filename)
    
    def get_image(self, id):
        if id is None:
            id = 'None'
        
        assert id in self.img_dict, 'get_image: id('+str(id)+') not found'
        return self.img_dict[id]
    
    # returns a deep copy of the desired animation set
    def get_animation_set(self, id):
        assert id in self.as_dict, 'get_animation_set: id('+id+') not found'
        as_copy = copy.deepcopy(self.as_dict[id])
        return as_copy
    
    ###########################
    #def get_animation(self, id, state):
    #    anim_set = self.get_animation_set(id)
    #    return anim_set[state]


class BackgroundRenderer(object):
    
    def __init__(self, cp_size, tilemap, id_man):
        super(BackgroundRenderer, self).__init__()
        self.size = tilemap.real_size
        self.id_manager = id_man
        # render tilemap to original_bg
        self.original_bg = self.create_background(tilemap)
        # buffer is the real background on which all the stuff gets rendered
        self.buffer = self.create_buffer()
        self.clipping_plane = self.create_clipping_plane(cp_size)
        self.adjust_cp_size()
    
    def create_clipping_plane(self, cp_size):
        w, h = self.size
        cp_w, cp_h = cp_size
        
        if cp_w > w:
            cp_w = w
        if cp_h > h:
            cp_h = h
        
        return pygame.Rect((0, 0), (cp_w, cp_h))
    
    # clipping plane shouldn't be bigger than the background
    def adjust_cp_size(self):
        w, h = self.size
        cp_w, cp_h = self.clipping_plane.size
        if cp_w > w:
            cp_w = w
        if cp_h > h:
            cp_h = h
    
    def create_background(self, tilemap):
        bg = pygame.Surface(self.size).convert()
        
        num_x, num_y = tilemap.size
        for y in range(num_y):
            for x in range(num_x):
                tile = tilemap.get_tile((x, y))

                tile_img = self.id_manager.get_image(tile.id)
                pos = tilemap.tile_index_to_coords((x, y))
                bg.blit(tile_img, pos)
        
        return bg
    
    def adjust_cp(self):
        w, h = self.size
        
        # adjust horizontal position
        if self.clipping_plane.left < 0:
            self.clipping_plane.left = 0
        elif self.clipping_plane.right > w:
            self.clipping_plane.right = w
        # adjust vertical position
        if self.clipping_plane.top < 0:
            self.clipping_plane.top = 0
        elif self.clipping_plane.bottom > h:
            self.clipping_plane.bottom = h
    
    # pos will be the center of the clipping_plane
    def view(self, pos):
        self.clipping_plane.center = pos
        self.adjust_cp()
    
    def create_buffer(self):
        return self.original_bg.copy()
    
    # update must be called before render and before anything
    # gets drawn to Background.buffer
    # this function blits (parts of the) the original_bg to buffer
    def update(self):
        self.buffer.blit(self.original_bg, self.clipping_plane,
                        self.clipping_plane)
    
    # render to screen
    def render(self, dest, pos):
        dest.blit(self.buffer, pos, self.clipping_plane)


class TargetRenderer(object):
    
    def __init__(self, target, id_man):
        super(TargetRenderer, self).__init__()
        self.target = target
        self.id_manager = id_man
        self.anim_set = self.id_manager.get_animation_set(target.id)
        
        self.prev_state = self.target.state
    
    def target_is_item(self):
        return isinstance(self.target, soaplogic.Item)
    
    def target_is_active(self):
        if self.target_is_item():
            if self.target.taken: # item has been taken by player
                return False
            else: # item is still available
                return True
        else: # target = Figure -> draw/update in any case
            return True
    
    def update(self):
        if self.target_is_active():
            anim = self.get_current_animation()
            # check if target state has changed
            # if yes, reset animation (frame index -> 0)
            if self.target.state != self.prev_state:
                anim.reset()
                self.prev_state = self.target.state
            anim.update()
    
    def render(self, dest):
        if self.target_is_active():
            image = self.get_current_frame()
            dest.blit(image, self.target.pos)
        
    def get_current_animation(self):
        return self.anim_set[self.target.state]
        
    def get_current_frame(self):
        anim = self.get_current_animation()
        frame_id = anim.get_current_frame_id()
        return self.id_manager.get_image(frame_id)


class GameRenderer(object):
    
    def __init__(self, cp_size, game, id_man):
        super(GameRenderer, self).__init__()
        self.game = game
        self.id_manager = id_man
        self.bg = BackgroundRenderer(cp_size, self.game.map, self.id_manager)
        # renderer_list contains all renderers for instances of Figure & Item
        self.renderer_list = self.create_target_renderers()
    
    # check if target is (partially?) inside our clipping plane
    def is_visible(self, target):
        t_size = self.game.map.tile_size
        t_rect = pygame.Rect(target.pos, t_size)
        
        cp = self.bg.clipping_plane
        if cp.colliderect(t_rect):
            return True
        else:
            return False
    
    def create_target_renderers(self):
        renderer_list = []
        
        # player renderer
        if self.game.player:
            p_r = TargetRenderer(self.game.player, self.id_manager)
            renderer_list.append(p_r)
        
        # enemy renderers
        for enemy in self.game.enemies:
            e_r = TargetRenderer(enemy, self.id_manager)
            renderer_list.append(e_r)
        
        # item renderers
        for item in self.game.items:
            i_r = TargetRenderer(item, self.id_manager)
            renderer_list.append(i_r)
        
        return renderer_list

    def update_bg(self):
        self.bg.update()
        player = self.game.player
        self.bg.view(player.pos)
    
    def update_renderer_list(self):
        for renderer in self.renderer_list:
            renderer.update() 
    
    # call update before rendering
    def update(self):
        self.update_bg()
        self.update_renderer_list()
    
    def render_all_targets(self, dest):
        for renderer in self.renderer_list:
            renderer.render(dest)
    
    def render(self, dest, pos=(0, 0)):
        self.render_all_targets(self.bg.buffer)
        self.bg.render(dest, pos)


class Animation(object):
    
    def __init__(self):
        super(Animation, self).__init__()
        # frames is a list of ID's
        self.frames = []
        self.delay = None # in milliseconds
        self.index = 0
        self.time_passed = 0 # in milliseconds
    
    def add_frame_id(self, frame):
        self.frames.append(frame)
    
    def get_current_frame_id(self):
        return self.frames[self.index]
    
    def update(self):
        if len(self.frames) > 1:
            self.time_passed += soaplogic.time_passed_ms()
            if self.time_passed >= self.delay:
                self.time_passed -= self.delay
                self.index += 1
                if self.index >= len(self.frames):
                    self.reset()
        else:
            self.reset()
    
    def reset(self):
        self.index = 0


class MenuEntryRenderer(object):
    
    def __init__(self, entry, text_size, text_color):
        super(MenuEntryRenderer, self).__init__()
        self.entry = entry
        self.text_size = text_size
        self.text_color = text_color
        self.font = self.create_default_font()
        self.image = self.create_entry_image()
        self.rect = self.image.get_rect()
    
    def create_default_font(self):
        return pygame.font.Font(None, self.text_size) 
    
    def get_pos(self):
        return self.rect.center
    
    def set_pos(self, pos):
        self.rect.center = pos
    
    pos = property(get_pos, set_pos)
    
    def create_entry_image(self):
        return self.font.render(self.entry.text, True, self.text_color)
    
    def render(self, dest):
        dest.blit(self.image, self.rect)


class MenuRenderer(object):
    
    def __init__(self, menu, menu_size, text_size, text_color, bg_color):
        super(MenuRenderer, self).__init__()
        self.menu = menu
        self.size = menu_size
        self.text_size = text_size
        self.text_color = text_color
        self.bg_color = bg_color
        self.line_pitch = self.get_line_pitch()
        self.menu_bg = self.create_menu_bg()
        self.bg_rect = self.menu_bg.get_rect()
        self.renderer_list = self.create_entry_renderers()
        self.render_entries_to_bg()
    
    def create_entry_renderers(self):
        renderer_list = []
        for entry in self.menu.entries:
            me_r = MenuEntryRenderer(entry, self.text_size, self.text_color)
            renderer_list.append(me_r)
        return renderer_list
    
    def calc_text_area(self):
        max_width = 0
        max_height = self.line_pitch
        
        for me_r in self.renderer_list:
            if me_r.rect.w > max_width:
                max_width = me_r.rect.w
            max_height += me_r.rect.h
            max_height += self.line_pitch
        
        # x, y coordinates are relative to self.menu_bg
        x = self.bg_rect.w / 2
        y = self.bg_rect.h / 2
        text_area = pygame.Rect((0, 0), (max_width, max_height))
        text_area.center = (x, y)
        
        return text_area
    
    def adjust_entry_positions(self):
        text_area = self.calc_text_area()
        p = self.line_pitch
        
        for i, me_r in enumerate(self.renderer_list):
            top_y = text_area.top + p + i * (me_r.rect.h + p)
            
            me_r.rect.centerx = text_area.centerx
            me_r.rect.top = top_y
        
    def create_menu_bg(self):
        bg = pygame.Surface(self.size)
        bg.convert_alpha()
        if len(self.bg_color) == 4:
            alpha = self.bg_color[3]
            bg.set_alpha(alpha)
        bg.fill(self.bg_color)
        r = pygame.Rect((0, 0), self.size)
        pygame.draw.rect(bg, self.text_color, r, 1)
        
        return bg
    
    def render_entries_to_bg(self):
        self.adjust_entry_positions()
        for me_r in self.renderer_list:
            me_r.render(self.menu_bg)
    
    def render_active_ellipse(self, dest):
        active_entry = self.menu.active_entry
        me_r = self.renderer_list[active_entry]
        
        # transform from bg_screen coordinates to screen coordinates
        x = self.bg_rect.left + me_r.rect.centerx
        y = self.bg_rect.top + me_r.rect.centery
        
        r = me_r.rect.copy()
        
        # scale rect: +50%
        r.w *= 1.5
        r.h *= 1.5
        r.center = (x, y)
        
        pygame.draw.ellipse(dest, self.text_color, r, 2)
    
    def render(self, dest):
        d_r = dest.get_rect()
        # draw the menu to the middle of the screen
        self.bg_rect.center = d_r.center
        dest.blit(self.menu_bg, self.bg_rect)
        
        self.render_active_ellipse(dest)
        
    
    def get_line_pitch(self):
        return self.text_size / 2
        
